/*
    v0.01 - initial



    up/down arrows
        - when live     - temporarily change to another stream on the same mux
                          without destroying the timeshift buffer
        - in playback   - select another stream in the recording

    left/right arrows
        - in playback   - select the pip channel from another stream in the recording

    record
        - menu

--todo:
-- mapping between recording slots and emma slots sometimes fails - fixed?
-- radio?
-- dont allow up/down in tempplaying when no extra channels added
-- nice to know what streams are available from playing
-- beware when switching from timeshift to rec; does it keep all the pids? - no
-- beware of delay from changing channels to f/w starting timeshift
   f/w will delete all the sr pids ... so, need to re-add
-- left/right arrows get a bit confused as to what chell is selected on sub window
-- add option to always timeshift mux
-- add option to remove steams from recording/timeshift
-- option to record pmt_pids
-- etc

*/
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"

#ifdef _WIN32
#else
#define ID_USERTAP 0xbdb02901
TAP_ID( ID_USERTAP );
TAP_PROGRAM_NAME("record-a-mux");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("v0.01 record a whole mux");
TAP_ETCINFO(__DATE__);
#endif

//----------------------------------------------------------------------------
extern void two_second_update();

extern void dump_memory();
extern void dump_play_info(type_play_info *pi);
extern void dump_rec_info(type_rec_info *ri);
extern void dump_ci(type_channel_info *ci);

//----------------------------------------------------------------------------
int playing_chNum = 0;
int sub_chNum = 0;

typedef struct{
    int recType;            //0=none, 1=rec, 2=timeshift
    int svcType;
    int slot;
    int channel_number;
} type_srpids;
type_srpids srpid[32];

int rmenu[0x20] = {0};
int tmenu[0x20] = {0};

word recording_status_rgn = 0;
int recording_status_showing = 0;
TYPE_Window recording_status_window;

word playing_status_rgn = 0;
int playing_status_showing = 0;
TYPE_Window playing_status_window;

int two_seconds_enabled = 1;
dword two_seconds = 0;


int TAP_Main(void);
int close(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
void oprintf(const void *fmt, ...);
int main_menu_handler(type_menu *menu, dword param1);
int record_menu_handler(type_menu *menu, dword param1);
int timeshift_menu_handler(type_menu *menu, dword param1);
int build_record_menu();
int build_timeshift_menu();
int key_handler(dword param1, dword param2);
int key_handler2(dword param1);
int key_handler3(dword param1);
int map_channel(int source);
int demap_channel(int channel);
int set_main_playback_pids(int channel, int video_pid, int audio_pid, int pcr_pid);
int set_sub_playback_pids(int channel, int video_pid, int audio_pid, int pcr_pid);
int get_next(int channel_type, int channel_number, int offset);
int add_this_channel(int slot, int recType, int channel_number);
int add_this_mux(int slot, int recType, int channel_number);
int add_pid(int pid, int slot0, int slot1, int recType, int svcType, int channel_number);
int on_record_stop(int slot);

//----------------------------------------------------------------------------
//menu stuff
char *tap_name;

type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {5, "xxx",    },

        {0, "Record ...", },
        {1, "Timeshift...", },
        {2, "Exit Tap", },
        {3, "Playback Status", },
        {4, "Recording Status", },
    },
};

type_menu record_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {5, "xxx",    },

        {0, "Record Status", },
        {1, "Record Channel", },
        {2, "Record Mux", },
        {1, "Add Mux to Recording", },
        {2, "Add Channel to Recording", },
    },
};

type_menu timeshift_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {3, "xxx",    },

        {0, "Timeshift Status", },
        {1, "Add Mux to Timeshift", },
        {2, "Add Channel to Timeshift", },
    },
};

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#ifdef _WIN32
#else
#include "TSRCommander.inc"

bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
#endif

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;
    int i;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }
    if ((fw_tshift_slot == 0) || (fw_rec_info == 0) || (fw_get_channel_info == 0)) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    ret = exTAP_Key_Extend2(1);
    if (ret < 0) TAP_Print("\nexTAP_Key_Extend2() failed, err=%d", ret);


    for (i=0; i<32; i++) {
        srpid[i].recType = RECTYPE_None;
        srpid[i].svcType = 0;
        srpid[i].slot = 0;
        srpid[i].channel_number = 0;
    }


    main_menu.visible = 0;
    record_menu.visible = 0;
    timeshift_menu.visible = 0;


    recording_status_rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
    if ((short)recording_status_rgn == -1) {
        recording_status_rgn = 0;
    }
    playing_status_rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
    if ((short)playing_status_rgn == -1) {
        playing_status_rgn = 0;
    }

    TSRCommanderInit(0, TRUE);
    return 1;                 //0=application mode, 1=tsr mode
}

int close(void)
{
    DeleteMenu(&main_menu);
    DeleteMenu(&record_menu);
    DeleteMenu(&timeshift_menu);

    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);

    exTAP_Key_Extend2(0);        //undo Key Mapping patch


    if (message_rgn) TAP_Osd_Delete(message_rgn);
    if (recording_status_rgn) TAP_Osd_Delete(recording_status_rgn);
    if (playing_status_rgn) TAP_Osd_Delete(playing_status_rgn);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    TAP_Exit();
    return 0;
}
//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    dword state, subState;
    int tick;

    switch (event) {
        case EVT_IDLE:
            tick = TAP_GetTick();
            //
            //message window
            //
            if (message_showing) {
                if (tick > message_showing) {
                    TAP_Osd_Delete(message_rgn);
                    message_rgn = 0;
                    message_showing = 0;
                }
            }

            //
            //housekeeping
            //
            if (two_seconds_enabled) {
                if (TAP_GetTick() > two_seconds) {
                    two_second_update();
                    two_seconds = TAP_GetTick() + (2*100);
                }
            }

            TSRCommanderWork();
            break;

        case EVT_KEY:
            TAP_GetState(&state, &subState);
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
//            TAP_Print("\nstate = %s, subState = %s", get_str(state_str, state), get_str(substate_str, subState));

            if (recording_status_showing) {
                param1 = key_handler2(param1);
                if (param1 == RKEY_Exit) {
            	    TAP_Win_Delete(&recording_status_window);
                    TAP_EnterNormal();
                    recording_status_showing = 0;
                    param1 = 0;
                    return param1;
                }
            }
            if (playing_status_showing) {
                param1 = key_handler3(param1);
                if (param1 == RKEY_Exit) {
            	    TAP_Win_Delete(&playing_status_window);
                    TAP_EnterNormal();
                    playing_status_showing = 0;
                    param1 = 0;
                    return param1;
                }
            }

            if (main_menu.visible) {
                param1 = main_menu_handler(&main_menu, param1);
            }
            else
            if (record_menu.visible) {
                param1 = record_menu_handler(&record_menu, param1);
            }
            else
            if (timeshift_menu.visible) {
                param1 = timeshift_menu_handler(&timeshift_menu, param1);
            }
            else {
                if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                    param1 = key_handler(param1, param2);
                }
            }
            break;

        default:
            break;
    }
    return param1;
}

//-----------------------------------------------------------------------------
//
//print multiline message to OSD
//  500=vertical position(0-576)
//  RGB(3,5,10)=background colour(dark blue)
//  200=hold time(2 seconds)
//  0=blocking
//
void oprintf(const void *fmt, ...)
{
    va_list argp;

    va_start(argp, fmt);
    vsprintf(strLog, fmt, argp);
//    TAP_Print("\n%s", strLog);
    ShowMessage_y(strLog, 500, RGB(3,5,10), 500, 0);
    va_end(argp);

    return;
}

//----------------------------------------------------------------------------
int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Record:
			DeleteMenu(menu);   //pass second record back to firmware
            return param1;
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:         // "Record ..."
                    build_record_menu();
                    ShowMenu(&record_menu);
				    return 0;
				case 1:         // "Timeshift ..."
                    build_timeshift_menu();
                    ShowMenu(&timeshift_menu);
				    return 0;
				case 2:         // "Close Tap"
                    close();
				    return 0;
				case 3:         // "Playback Status"
                    dump_playing_status();
				    return 0;
				case 4:         // "Recording Status"
                    dump_recording_status();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}

int record_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;
    type_channel_info ci;
    int slot;
    TYPE_RecInfo ri;
    byte dummy3[0x80];
    type_rec_info *rec_info;
    TYPE_TapChInfo chInfo;
    int channel_type, channel_number;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:         // "Record Status"
//				    record_status();
                    dump_recording_status();
				    return 0;
				case 1:         // "Record Channel"
                    ret = TAP_Hdd_StartRecord();
				    return 0;
				case 2:         // "Record Mux"
/*
                    TAP_Print("\nrecord this mux");
                    ret = TAP_Hdd_StartRecord();
                    if (ret == 0) return 0;
//!!! wait for recording to really start
//??? how
//--disable for now
				    //which slot ?
                    TAP_Channel_GetCurrent(&channel_type, &channel_number);
                    ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, playing_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                    for (slot=0; slot<2; slot++) {
                        rec_info = (type_rec_info *)(fw_rec_info+(0x88*slot));
                        TAP_Hdd_GetRecInfo(slot, &ri);
                        if (ri.recType == RECTYPE_Normal) {
//                            ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, ri.svcNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                            //does it match this mux
                            if (rec_info->transport_stream_id == ci.transport_stream_id) {
                                TAP_Print("\nadding to slot %x", slot);
            				    add_this_mux(slot, RECTYPE_Normal, playing_chNum);
			            	    return 0;
                            }
                        }
                    }
                    oprintf("Cannot Add: Recording not running");
*/
                    oprintf("option disabled - start record; wait 10 seconds; Add Mux to Recording");
				    return 0;
				case 3:         // "Add Mux to Recording"
				    //which slot ?
                    TAP_Channel_GetCurrent(&channel_type, &channel_number);
                    ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, playing_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                    for (slot=0; slot<2; slot++) {
                        rec_info = (type_rec_info *)(fw_rec_info+(0x88*slot));
                        TAP_Hdd_GetRecInfo(slot, &ri);
                        if (ri.recType == RECTYPE_Normal) {
//                            ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, ri.svcNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                            //does it match this mux
                            if (rec_info->transport_stream_id == ci.transport_stream_id) {
                                TAP_Print("\nadding to slot %x", slot);
            				    add_this_mux(slot, RECTYPE_Normal, playing_chNum);
			            	    return 0;
                            }
                        }
                    }
                    oprintf("Cannot Add: Recording not running");
				    return 0;
				case 4:         // "Add Channel to Recording"
				case 5:         // "Add Channel to Recording"
				case 6:         // "Add Channel to Recording"
				case 7:         // "Add Channel to Recording"
				case 8:         // "Add Channel to Recording"
				case 9:         // "Add Channel to Recording"
				case 10:        // "Add Channel to Recording"
				case 11:        // "Add Channel to Recording"
				case 12:        // "Add Channel to Recording"
				case 13:        // "Add Channel to Recording"
				case 14:        // "Add Channel to Recording"
				case 15:        // "Add Channel to Recording"
				    //which slot ?
                    TAP_Channel_GetCurrent(&channel_type, &channel_number);
                    ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, playing_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                    for (slot=0; slot<2; slot++) {
                        rec_info = (type_rec_info *)(fw_rec_info+(0x88*slot));
                        TAP_Hdd_GetRecInfo(slot, &ri);
                        if (ri.recType == RECTYPE_Normal) {
                            //does it match this mux
                            if (rec_info->transport_stream_id == ci.transport_stream_id) {
                                TAP_Print("\nadding to slot %x", slot);
            				    ret = add_this_channel(slot, RECTYPE_Normal, rmenu[selection-4]);
                                if (ret == 0) {
                                    TAP_Channel_GetInfo(channel_type, rmenu[selection-4], &chInfo);
                                    oprintf("Adding %d, %s to Recording", chInfo.logicalChNum, chInfo.chName);
                                }
			            	    return 0;
                            }
                        }
                    }
                    oprintf("Cannot Add: Recording not running");
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}

int timeshift_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int i, ret;
    int slot;
    TYPE_TapChInfo chInfo;
    int channel_type, channel_number;
    int emma_pid, active, slot0, slot1, pid;
    TYPE_RecInfo ri;
    byte dummy3[0x80];

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:         // "Timeshift Status"
//				    timeshift_status();
                    dump_recording_status();
				    return 0;
				case 1:         // "Add Mux to Timeshift"
                    //don't add unless slot is timeshifting
                    slot = (int)*fw_tshift_slot;
                    TAP_Hdd_GetRecInfo(slot, &ri);
                    if (ri.recType != RECTYPE_Timeshift) {
                        oprintf("Cannot Add: Timeshift not running");
                        return 0;
                    }

				    //assume the current channel is already timeshifting
				    //get its slot
                    TAP_Channel_GetCurrent(&channel_type, &channel_number);
                    TAP_Channel_GetInfo(channel_type, channel_number, &chInfo);

                    //match slot number
                    for (i=0; i<32; i++) {
                        emma_pid = *(dword *)(EMMA_TTSD_SR_PID0+(i*4));
                        active = (emma_pid & 0x00020000) ? 1 : 0;
                        slot0 = (emma_pid & 0x00010000) ? 1 : 0;
                        slot1 = (emma_pid & 0x00002000) ? 1 : 0;
                        pid = emma_pid & 0x1fff;
                        if (active) {
                            if (pid == chInfo.pcrPid) {
                                slot = slot1;
                                break;
                            }
                        }
                    }
                    if (i == 32) {
                        oprintf("Cannot Add: Timeshift not running");
                        return 0;
                    }

				    add_this_mux(slot, RECTYPE_Timeshift, playing_chNum);
                    oprintf("Adding mux to Timeshift", chInfo.logicalChNum, chInfo.chName);
				    return 0;
				case 2:         // "Add Channel to Timeshift"
				case 3:         // "Add Channel to Timeshift"
				case 4:         // "Add Channel to Timeshift"
				case 5:         // "Add Channel to Timeshift"
				case 6:         // "Add Channel to Timeshift"
				case 7:         // "Add Channel to Timeshift"
				case 8:         // "Add Channel to Timeshift"
				case 9:         // "Add Channel to Timeshift"
				case 10:        // "Add Channel to Timeshift"
				case 11:        // "Add Channel to Timeshift"
				case 12:        // "Add Channel to Timeshift"
				case 13:        // "Add Channel to Timeshift"
                    //don't add unless slot is timeshifting
                    slot = (int)*fw_tshift_slot;
                    TAP_Hdd_GetRecInfo(slot, &ri);
                    if (ri.recType != RECTYPE_Timeshift) {
                        oprintf("Cannot Add: Timeshift not running");
                        return 0;
                    }

				    //assume the current channel is already timeshifting
				    //get its slot
                    TAP_Channel_GetCurrent(&channel_type, &channel_number);
                    TAP_Channel_GetInfo(channel_type, channel_number, &chInfo);

                    //match slot number
                    for (i=0; i<32; i++) {
                        emma_pid = *(dword *)(EMMA_TTSD_SR_PID0+(i*4));
                        active = (emma_pid & 0x00020000) ? 1 : 0;
                        slot0 = (emma_pid & 0x00010000) ? 1 : 0;
                        slot1 = (emma_pid & 0x00002000) ? 1 : 0;
                        pid = emma_pid & 0x1fff;
                        if (active) {
                            if (pid == chInfo.pcrPid) {
                                slot = slot1;
                                break;
                            }
                        }
                    }
                    if (i == 32) {
                        oprintf("Cannot Add: Timeshift not running");
                        return 0;
                    }

				    ret = add_this_channel(slot, RECTYPE_Timeshift, tmenu[selection-2]);
                    if (ret == 0) {
                        TAP_Channel_GetCurrent(&channel_type, &channel_number);
                        TAP_Channel_GetInfo(channel_type, tmenu[selection-2], &chInfo);
                        oprintf("Adding %d, %s to Timeshift", chInfo.logicalChNum, chInfo.chName);
                    }
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}


int build_record_menu()
{
    int ret;
    int i,j;
    TYPE_TapChInfo this_chInfo;
    TYPE_TapChInfo chInfo;
    int nTvSvc, nRadioSvc;
    int channel_type, channel_number;
    int num_channels;

	//which slot ?
    TAP_Channel_GetCurrent(&channel_type, &channel_number);
    TAP_Channel_GetInfo(channel_type, playing_chNum, &this_chInfo);

    //figure out all channels on the same mux
    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    num_channels = (channel_type == SVC_TYPE_Tv) ? nTvSvc : nRadioSvc;
    j = 0;
    for (i=0; i<num_channels; i++) {
        TAP_Channel_GetInfo(channel_type, i, &chInfo);
        if ((chInfo.tsId == this_chInfo.tsId) && ((chInfo.videoPid & 0x1fff) != 0) && ((chInfo.videoPid & 0x1fff) != 0x1fff)) {
//            TAP_Print("\nmatched tsid: %04x", chInfo.tsId);
//            strcpy(record_menu.item[5+j].str, "Add Channel to Recording");
            sprintf(record_menu.item[5+j].str, "Add %d. %s", chInfo.logicalChNum, chInfo.chName);
            rmenu[j] = i;
            j++;
        }
    }
    record_menu.item[0].id = 4+j;
}

int build_timeshift_menu()
{
    int ret;
    int i,j;
    TYPE_TapChInfo this_chInfo;
    TYPE_TapChInfo chInfo;
    int nTvSvc, nRadioSvc;
    int channel_number, channel_type;
    int num_channels;

	//which slot ?
    TAP_Channel_GetCurrent(&channel_type, &channel_number);
    TAP_Channel_GetInfo(channel_type, playing_chNum, &this_chInfo);

    //figure out all channels on the same mux
    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    num_channels = (channel_type == SVC_TYPE_Tv) ? nTvSvc : nRadioSvc;
    j = 0;
    for (i=0; i<num_channels; i++) {
        TAP_Channel_GetInfo(channel_type, i, &chInfo);
        if ((chInfo.tsId == this_chInfo.tsId) && ((chInfo.videoPid & 0x1fff) != 0) && ((chInfo.videoPid & 0x1fff) != 0x1fff)) {
//            TAP_Print("\nmatched tsid: %04x", chInfo.tsId);
//            strcpy(record_menu.item[3+j].str, "Add Channel to Timeshift");
            sprintf(timeshift_menu.item[3+j].str, "Add %d. %s", chInfo.logicalChNum, chInfo.chName);
            tmenu[j] = i;
            j++;
        }
    }
    timeshift_menu.item[0].id = 2+j;
}

//----------------------------------------------------------------------------
int key_handler(dword param1, dword param2)
{
    int ret;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    int slot;
    TYPE_RecInfo ri;
    byte dummy3[0x80];
    type_rec_info *rec_info;
    type_channel_info ci;
    int channel_type, channel_number;
    TYPE_TapChInfo this_chInfo;
    TYPE_TapChInfo chInfo;
    dword channel;
    int source;
    int main_tuner, sub_tuner;

    main_tuner = TAP_GetTuner(CHANNEL_Main);
    sub_tuner = TAP_GetTuner(CHANNEL_Sub);
    ret = TAP_Hdd_GetPlayInfo(&pi);


    switch(param1) {
        case RKEY_Record:
            ShowMenu(&main_menu);
            return 0;
            break;

        case RKEY_Up:
            if (param2 == exRKEY_ChUp) {
            }
            else {
                //change the main

                //find the next channel on this mux
                TAP_Channel_GetCurrent(&channel_type, &channel_number);
                playing_chNum = get_next(channel_type, playing_chNum, 1);
                TAP_Channel_GetInfo(channel_type, playing_chNum, &chInfo);
                TAP_Print("\nChanging Main to: %5s - %3d - %s", get_str(service_type_str, channel_type), chInfo.logicalChNum, chInfo.chName);
                oprintf("%3d - %s", chInfo.logicalChNum, chInfo.chName);

                ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, playing_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                //dump_ci(&ci);

                source = (pi.playMode == PLAYMODE_None) ? main_tuner : 3;
                channel = map_channel(source);
                set_main_playback_pids(channel, ci.video_pid, ci.audio_pid, ci.pcr_pid);
                return 0;
            }
            break;
        case RKEY_Down:
            if (param2 == exRKEY_ChDown) {
            }
            else {
                //change the main

                //find the previous channel on this mux
                TAP_Channel_GetCurrent(&channel_type, &channel_number);
                playing_chNum = get_next(channel_type, playing_chNum, -1);
                TAP_Channel_GetInfo(channel_type, playing_chNum, &chInfo);
                TAP_Print("\nChanging Main to: %5s - %3d - %s", get_str(service_type_str, channel_type), chInfo.logicalChNum, chInfo.chName);
                oprintf("%3d - %s", chInfo.logicalChNum, chInfo.chName);

                ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, playing_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                //dump_ci(&ci);

                source = (pi.playMode == PLAYMODE_None) ? main_tuner : 3;
                channel = map_channel(source);
                set_main_playback_pids(channel, ci.video_pid, ci.audio_pid, ci.pcr_pid);
                return 0;
            }
            break;

        case RKEY_Right:
            if (param2 == exRKEY_VolUp) {
            }
            else {
//                if (pi.playMode != PLAYMODE_None) {
                    //change the sub
                    //if sub_chNum not on this mux, use playing_chNum
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, playing_chNum, &chInfo);
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_chNum, &this_chInfo);
                    if (this_chInfo.tsId != chInfo.tsId) {
                        sub_chNum = playing_chNum;
                    }

                    //find the next channel on this mux
                    sub_chNum = get_next(SVC_TYPE_Tv, sub_chNum, 1);
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_chNum, &chInfo);
                    TAP_Print("\nChanging Sub to: %5s - %3d - %s", get_str(service_type_str, SVC_TYPE_Tv), chInfo.logicalChNum, chInfo.chName);
                    oprintf("SUB: %3d - %s", chInfo.logicalChNum, chInfo.chName);

                    ret = exTAP_CallFirmware(fw_get_channel_info, SVC_TYPE_Tv, sub_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                    //dump_ci(&ci);

                    //use the main tuner
//                    source = (pi.playMode == PLAYMODE_None) ? sub_tuner : 3;
                    source = (pi.playMode == PLAYMODE_None) ? main_tuner : 3;
                    channel = map_channel(source);
                    set_sub_playback_pids(channel, ci.video_pid, ci.audio_pid, ci.pcr_pid);
//                }
                return 0;
            }
            break;
        case RKEY_Left:
            if (param2 == exRKEY_VolDown) {
            }
            else {
//                if (pi.playMode != PLAYMODE_None) {
                    //if sub_chNum not on this mux, use playing_chNum
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, playing_chNum, &chInfo);
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_chNum, &this_chInfo);
                    if (this_chInfo.tsId != chInfo.tsId) {
                        sub_chNum = playing_chNum;
                    }
                    //change the sub
                    //find the previous channel on this mux
                    sub_chNum = get_next(SVC_TYPE_Tv, sub_chNum, -1);
                    TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_chNum, &chInfo);
                    TAP_Print("\nChanging Sub to: %5s - %3d - %s", get_str(service_type_str, SVC_TYPE_Tv), chInfo.logicalChNum, chInfo.chName);
                    oprintf("SUB: %3d - %s", chInfo.logicalChNum, chInfo.chName);

                    ret = exTAP_CallFirmware(fw_get_channel_info, SVC_TYPE_Tv, sub_chNum, (dword)&ci,  0,0,0,0,0,0,0,0);
                    //dump_ci(&ci);

                    //use the main tuner
//                    source = (pi.playMode == PLAYMODE_None) ? sub_tuner : 3;
                    source = (pi.playMode == PLAYMODE_None) ? main_tuner : 3;
                    channel = map_channel(source);
                    set_sub_playback_pids(channel, ci.video_pid, ci.audio_pid, ci.pcr_pid);
//                }
                return 0;
            }
            break;
    }
    return param1;
}

int key_handler2(dword param1)
{
    switch(param1) {
        case RKEY_Up:
             TAP_Win_Action(&recording_status_window, RKEY_Up);
             return 0;
             break;
        case RKEY_Down:
            TAP_Win_Action(&recording_status_window, RKEY_Down);
            return 0;
            break;
    }
    return param1;
}

int key_handler3(dword param1)
{
    switch(param1) {
        case RKEY_Up:
             TAP_Win_Action(&playing_status_window, RKEY_Up);
             return 0;
             break;
        case RKEY_Down:
            TAP_Win_Action(&playing_status_window, RKEY_Down);
            return 0;
            break;
    }
    return param1;
}

//-----------------------------------------------------------------------------
int map_channel(int source)
{
    int channel;

    channel = -1;
    if (source == 0) {
        if ((*(dword *)EMMA_TTSD_CHA_CNT & 0xf0) == 0xc0) {  //t0
            channel = 0;  //ch_a
        }
        else
        if ((*(dword *)EMMA_TTSD_CHB_CNT & 0xf0) == 0xc0) {  //t0
            channel = 1;  //ch_b
        }
        else
        if ((*(dword *)EMMA_TTSD_CHC_CNT & 0xf0) == 0xc0) {  //t0
            channel = 2;  //ch_c
        }
    }
    else
    if (source == 1) {
        if ((*(dword *)EMMA_TTSD_CHA_CNT & 0xf0) == 0xe0) {  //t1
            channel = 0;  //ch_a
        }
        else
        if ((*(dword *)EMMA_TTSD_CHB_CNT & 0xf0) == 0xe0) {  //t1
            channel = 1;  //ch_b
        }
        else
        if ((*(dword *)EMMA_TTSD_CHC_CNT & 0xf0) == 0xe0) {  //t1
            channel = 2;  //ch_c
        }
    }
    else {
        if ((*(dword *)EMMA_TTSD_CHA_CNT & 0xf0) == 0xb0) {  //playback
            channel = 0;  //ch_a
        }
        else
        if ((*(dword *)EMMA_TTSD_CHB_CNT & 0xf0) == 0xb0) {  //playback
            channel = 1;  //ch_b
        }
        else
        if ((*(dword *)EMMA_TTSD_CHC_CNT & 0xf0) == 0xb0) {  //playback
            channel = 2;  //ch_c
        }
    }
//    TAP_Print("\nsource=%x, channel=%x", source, channel);
    return channel;
}

int demap_channel(int channel)
{
    int source;

    source = -1;
    if (channel == 0) {
        if ((*(dword *)EMMA_TTSD_CHA_CNT & 0xf0) == 0xc0) {  //t0
            source = 0;
        }
        else
        if ((*(dword *)EMMA_TTSD_CHA_CNT & 0xf0) == 0xe0) {  //t0
            source = 1;
        }
        else
        if ((*(dword *)EMMA_TTSD_CHA_CNT & 0xf0) == 0xb0) {  //t0
            source = 2;
        }
    }
    else
    if (channel == 1) {
        if ((*(dword *)EMMA_TTSD_CHB_CNT & 0xf0) == 0xc0) {  //t0
            source = 0;
        }
        else
        if ((*(dword *)EMMA_TTSD_CHB_CNT & 0xf0) == 0xe0) {  //t1
            source = 1;
        }
        else
        if ((*(dword *)EMMA_TTSD_CHB_CNT & 0xf0) == 0xb0) {  //playback
            source = 2;
        }
    }
    else {
        if ((*(dword *)EMMA_TTSD_CHC_CNT & 0xf0) == 0xc0) {  //t0
            source = 0;
        }
        else
        if ((*(dword *)EMMA_TTSD_CHC_CNT & 0xf0) == 0xe0) {  //t1
            source = 1;
        }
        else
        if ((*(dword *)EMMA_TTSD_CHC_CNT & 0xf0) == 0xb0) {  //playback
            source = 2;
        }
    }
//    TAP_Print("\nchannel=%x, source=%x", channel, source);
    return source;
}

//
//
int set_main_playback_pids(int channel, int video_pid, int audio_pid, int pcr_pid)
{
    dword emma_channel;

    //
    //emma has 3 channels, a,b,c
    //channel a-0x00000000, b-0x20000000, c-0x40000000
    //
    emma_channel = (channel == 0) ? 0x00000000
                 : (channel == 1) ? 0x20000000
                 : (channel == 2) ? 0x40000000 : 0;

    //an interrupt gets generated after changing the pcr, which puts it back
    //to the value in this location, so change the location first
    *fw_pcr1_pid = emma_channel | 0x000 | (pcr_pid & 0x1fff);
    *(dword *)(EMMA_TTSD_1ST_PCR_PID)   = emma_channel | 0x000 | (pcr_pid & 0x1fff);
    *(dword *)(EMMA_TTSD_1ST_VIDEO_PID) = emma_channel | 0x4000 | (video_pid & 0x1fff);
    *(dword *)(EMMA_TTSD_1ST_AUDIO_PID) = emma_channel | 0x4000 | (audio_pid & 0x1fff);

//    TAP_Print("\nEMMA_TTSD_1ST_PCR_PID   = %x", *(dword *)(EMMA_TTSD_1ST_PCR_PID));
//    TAP_Print("\nEMMA_TTSD_1ST_VIDEO_PID = %x", *(dword *)(EMMA_TTSD_1ST_VIDEO_PID));
//    TAP_Print("\nEMMA_TTSD_1ST_AUDIO_PID = %x", *(dword *)(EMMA_TTSD_1ST_AUDIO_PID));
}

int set_sub_playback_pids(int channel, int video_pid, int audio_pid, int pcr_pid)
{
    dword emma_channel;
    dword *fw_pcr2_pid = fw_pcr1_pid + 2;
    //
    //emma has 3 channels, a,b,c
    //channel a-0x00000000, b-0x20000000, c-0x40000000
    //
    emma_channel = (channel == 0) ? 0x00000000
                 : (channel == 1) ? 0x20000000
                 : (channel == 2) ? 0x40000000 : 0;

    *fw_pcr2_pid = emma_channel | 0x000 | (pcr_pid & 0x1fff);    //necessary?
    *(dword *)(EMMA_TTSD_2ND_PCR_PID)   = emma_channel | 0x000 | (pcr_pid & 0x1fff);
    *(dword *)(EMMA_TTSD_2ND_VIDEO_PID) = emma_channel | 0x4000 | (video_pid & 0x1fff);
//    *(dword *)(EMMA_TTSD_2ND_AUDIO_PID) = emma_channel | 0x4000 | (audio_pid & 0x1fff);

//    TAP_Print("\nEMMA_TTSD_2ND_PCR_PID   = %x", *(dword *)(EMMA_TTSD_2ND_PCR_PID));
//    TAP_Print("\nEMMA_TTSD_2ND_VIDEO_PID = %x", *(dword *)(EMMA_TTSD_2ND_VIDEO_PID));
//    TAP_Print("\nEMMA_TTSD_2ND_AUDIO_PID = %x", *(dword *)(EMMA_TTSD_2ND_AUDIO_PID));
}


//
//find the next/previous channel on this mux
int get_next(int channel_type, int channel_number, int offset)
{
    int nTvSvc, nRadioSvc;
    int num_channels;
    int ret;
    type_channel_info ci;
    int playing_tsid;
    int next_channel_number;

    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    num_channels = (channel_type == SVC_TYPE_Tv) ? nTvSvc : nRadioSvc;

    //get the playing status
    ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, channel_number, (dword)&ci,  0,0,0,0,0,0,0,0);
    playing_tsid = ci.transport_stream_id;

    next_channel_number = channel_number;
    next_channel_number = next_channel_number + offset;
    while (next_channel_number != channel_number) {
        if (next_channel_number == num_channels) next_channel_number = 0;
        if (next_channel_number < 0) next_channel_number = num_channels - 1;

        ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, next_channel_number, (dword)&ci,  0,0,0,0,0,0,0,0);
        //stop on the next channel that has video
        if ((ci.transport_stream_id == playing_tsid) && (ci.video_pid & 0x1fff)  && ((ci.video_pid & 0x1fff) != 0x1fff)) {
//            TAP_Print("\nnext_channel_number=%x, ci.transport_stream_id=%x, playing_tsid=%x, ci.video_pid=%x", next_channel_number, ci.transport_stream_id, playing_tsid, ci.video_pid);
            break;
        }
        next_channel_number = next_channel_number + offset;
    }
    return next_channel_number;
}

//-----------------------------------------------------------------------------
int add_this_channel(int slot, int recType, int channel_number)
{
    int ret;
    int channel_type, channel_number1;
    TYPE_TapChInfo  channel_info;
    type_channel_info ci;
    type_rec_info *rec_info;

    TAP_Print("\nadd this channel, slot=%x", slot);

    //get the current channel
//    ShowMainChannelInfo();
    TAP_Channel_GetCurrent(&channel_type, &channel_number1);
    TAP_Channel_GetInfo(channel_type, channel_number, &channel_info);
    ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, channel_number, (dword)&ci,  0,0,0,0,0,0,0,0);
//    TAP_Print("\nservice_type = %x", channel_type);
//    TAP_Print("\nchannel_num  = %x", channel_number);
//    dump_ci(&this_ci);

    //get the current slot
    rec_info = (type_rec_info *)(fw_rec_info+(0x88*slot));
    if (rec_info->transport_stream_id != ci.transport_stream_id) {
        TAP_Print("\noops, tsid: current = %x, slot %x = %x", ci.transport_stream_id, slot, rec_info->transport_stream_id);
        return -1;
    }

    //figure out all the pids needed
    if (ci.video_pid & 0x1fff) {
        ret = add_pid(ci.video_pid, slot, slot, recType, channel_type, channel_number);
        ret = add_pid(ci.audio_pid, slot, slot, recType, channel_type, channel_number);
        ret = add_pid(ci.pcr_pid,   slot, slot, recType, channel_type, channel_number);
//      ret = add_pid(ci.pmt_pid,   slot, slot, recType, channel_type, channel_number);
    }
    return 0;
}

int add_this_mux(int slot, int recType, int channel_number)
{
    int ret;
    int channel_type, channel_number1;
    TYPE_TapChInfo  channel_info;
    type_channel_info this_ci;
    type_channel_info ci;
    type_rec_info *rec_info;

    int nTvSvc, nRadioSvc;
    int num_channels;
    int i;

    TAP_Print("\nadd this mux, slot=%x", slot);

    //get the current channel
//    ShowMainChannelInfo();
    TAP_Channel_GetCurrent(&channel_type, &channel_number1);
    TAP_Channel_GetInfo(channel_type, channel_number, &channel_info);
    ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, channel_number, (dword)&this_ci,  0,0,0,0,0,0,0,0);
//    TAP_Print("\nservice_type = %x", channel_type);
//    TAP_Print("\nchannel_num  = %x", channel_number);
//    dump_ci(&this_ci);

    //get the current slot
    rec_info = (type_rec_info *)(fw_rec_info+(0x88*slot));
    if (rec_info->transport_stream_id != this_ci.transport_stream_id) {
        TAP_Print("\noops, tsid: current = %x, slot %x = %x", this_ci.transport_stream_id, slot, rec_info->transport_stream_id);
        return 0;
    }

    //figure out all the pids needed
    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    num_channels = (channel_type == SVC_TYPE_Tv) ? nTvSvc : nRadioSvc;
    for (i=0; i<num_channels; i++) {
        ret = exTAP_CallFirmware(fw_get_channel_info, channel_type, i, (dword)&ci,  0,0,0,0,0,0,0,0);
        if (ci.transport_stream_id == this_ci.transport_stream_id) {
//            TAP_Print("\nmatched tsid: %04x", ci.transport_stream_id);
//            dump_ci(&ci);

            if (ci.video_pid & 0x1fff) {
                ret = add_pid(ci.video_pid, slot, slot, recType, channel_type, i);
                ret = add_pid(ci.audio_pid, slot, slot, recType, channel_type, i);
                ret = add_pid(ci.pcr_pid,   slot, slot, recType, channel_type, i);
//              ret = add_pid(ci.pmt_pid,   slot, slot, recType, channel_type, i);
            }
        }
    }
}



//find a free sr pid, add the video/audio/pcr pids
int add_pid(int pid, int slot0, int slot1, int recType, int svcType, int channel_number)
{
    int i;
    dword emma_pid;

    if ((pid & 0x1fff) == 0) return -2;

    //is this pid already recording?
    for (i=0; i<32; i++) {
        emma_pid = *(dword *)(EMMA_TTSD_SR_PID0+(i*4));
        if (emma_pid & 0x00020000) {                                    //active
            if ((emma_pid & 0x00010000) == ((slot0 & 1) << 16)) {       //slot0
                if ((emma_pid & 0x00002000) == ((slot1 & 1) << 13)) {   //slot1
                    if ((emma_pid & 0x1fff) == (pid & 0x1fff)) {        //pid
//                        TAP_Print("\nalready got: i=%d, %08x", i, *(dword *)(EMMA_TTSD_SR_PID0+(i*4)));
                        return i;
                    }
                }
            }
        }
    }


    if (slot1) {
        //8..15, 24..31
        for (i=31; i>=24; i--) {
            if ((*(dword *)(EMMA_TTSD_SR_PID0+(i*4)) & 0x00020000) == 0) {
//                TAP_Print("\nadding: i=%d, %08x", i, 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff));
                *(dword *)(EMMA_TTSD_SR_PID0+(i*4)) = 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff);

                //remember that we have set these pids
                srpid[i].recType = recType;
                srpid[i].svcType = svcType;
                srpid[i].slot = slot1;
                srpid[i].channel_number = channel_number;
                return i;
            }
        }
        for (i=15; i>=8; i--) {
            if ((*(dword *)(EMMA_TTSD_SR_PID0+(i*4)) & 0x00020000) == 0) {
//                TAP_Print("\nadding: i=%d, %08x", i, 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff));
                *(dword *)(EMMA_TTSD_SR_PID0+(i*4)) = 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff);

                //remember that we have set these pids
                srpid[i].recType = recType;
                srpid[i].svcType = svcType;
                srpid[i].slot = slot1;
                srpid[i].channel_number = channel_number;
                return i;
            }
        }
    }
    else {
        //0..7, 16..23
        for (i=23; i>=16; i--) {
            if ((*(dword *)(EMMA_TTSD_SR_PID0+(i*4)) & 0x00020000) == 0) {
//                TAP_Print("\nadding: i=%d, %08x", i, 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff));
                *(dword *)(EMMA_TTSD_SR_PID0+(i*4)) = 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff);

                //remember that we have set these pids
                srpid[i].recType = recType;
                srpid[i].svcType = svcType;
                srpid[i].slot = slot1;
                srpid[i].channel_number = channel_number;
                return i;
            }
        }
        for (i=7; i>=0; i--) {
            if ((*(dword *)(EMMA_TTSD_SR_PID0+(i*4)) & 0x00020000) == 0) {
//                TAP_Print("\nadding: i=%d, %08x", i, 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff));
                *(dword *)(EMMA_TTSD_SR_PID0+(i*4)) = 0x00020000 | ((slot0 & 1) << 16) | ((slot1 & 1) << 13) | (pid & 0x1fff);

                //remember that we have set these pids
                srpid[i].recType = recType;
                srpid[i].svcType = svcType;
                srpid[i].slot = slot1;
                srpid[i].channel_number = channel_number;
                return i;
            }
        }
    }
    //no spare slots
    return -1;
}

int on_record_stop(int slot)
{
    int i;

    //stop all the excess pids
    for (i=0; i<32; i++) {
        if (srpid[i].slot == slot) {
//            TAP_Print("\nslot %x, stopping %d, %x", slot, i, *(dword *)(EMMA_TTSD_SR_PID0+(i*4)));
            *(dword *)(EMMA_TTSD_SR_PID0+(i*4)) = 0;
            srpid[i].recType = RECTYPE_None;
            srpid[i].svcType = 0;
            srpid[i].slot = 0;
            srpid[i].channel_number = 0;
        }
    }

}

