
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include <tap.h>
#include "exTap.h"
#include "tf5800.h"

//----------------------------------------------------------------------------
extern int playing_chNum;
extern int sub_chNum;

typedef struct{
    int recType;            //0=none, 1=rec, 2=timeshift
    int svcType;
    int slot;
    int channel_number;
} type_srpids;
extern type_srpids srpid[32];

extern word recording_status_rgn;
extern int recording_status_showing;
extern TYPE_Window recording_status_window;

extern word playing_status_rgn;
extern int playing_status_showing;
extern TYPE_Window playing_status_window;

//----------------------------------------------------------------------------
void dump_memory();
void dump_play_info(type_play_info *pi);
void dump_rec_info(type_rec_info *ri);
void dump_ci(type_channel_info *ci);
int dump_recording_status();
int dump_playing_status();

//----------------------------------------------------------------------------
void dump_memory()
{
    dword heapSize, freeHeapSize, availHeapSize;

    TAP_MemInfo(&heapSize, &freeHeapSize, &availHeapSize);
    oprintf("heapSize=%d        freeHeapSize=%d        availHeapSize=%d", heapSize, freeHeapSize, availHeapSize);
}

//----------------------------------------------------------------------------
void dump_play_info(type_play_info *pi)
{
    TAP_Print("\nplay_info()");
    TAP_Print("\nplayMode     = %02x, %s", pi->playMode, get_str(playmode_str, pi->playMode));   //03, 02
    TAP_Print("\nbyte_01      = %02x", pi->byte_01);    //00
    TAP_Print("\nbyte_02      = %02x", pi->byte_02);    //00
    TAP_Print("\nbyte_03      = %02x", pi->byte_03);    //00, 03
    TAP_Print("\ntrickMode    = %x, %s", pi->trickMode, get_str(trickmode_str, pi->trickMode));  //00000000
    TAP_Print("\nspeed        = %x",   pi->speed);      //00000001
    TAP_Print("\ndword_0c     = %08x", pi->dword_0c);   //00000001

    TAP_Print("\nstart time   = %s", mjdToString(pi->start_time));
    TAP_Print("\ndword_14     = %08x", pi->dword_14);   //03000000
    TAP_Print("\ndword_18     = %08x", pi->dword_18);   //00005f2f
    TAP_Print("\ndword_1c     = %08x", pi->dword_1c);   //000000b0
    TAP_Print("\ndword_20     = %08x", pi->dword_20);   //001be800
    TAP_Print("\nfilename     = %s",   pi->filename);
    TAP_Print("\nchannel_name = %s", pi->channel_name);
    TAP_Print("\ndword_80     = %08x", pi->dword_80);   //00000080
    TAP_Print("\ndword_84     = %08x", pi->dword_84);   //00000000
    TAP_Print("\ndword_88     = %08x", pi->dword_88);   //00000000
    TAP_Print("\ndword_8c     = %08x", pi->dword_8c);   //22000000

    TAP_Print("\ndword_90     = %08x", pi->dword_90);   //82018edc
    TAP_Print("\ndword_94     = %08x", pi->dword_94);   //e812c6ae
    TAP_Print("\ndword_98     = %08x", pi->dword_98);   //00000000
    TAP_Print("\ndword_9c     = %08x", pi->dword_9c);   //18201800
    TAP_Print("\ndword_a0     = %08x", pi->dword_a0);   //00000000
    TAP_Print("\ndword_a4     = %08x", pi->dword_a4);   //d87d86b7
}

void dump_rec_info(type_rec_info *ri)
{
    TAP_Print("\nrec_info()");
    TAP_Print("\nrecType             = %02x, %s", ri->recType, get_str(rectype_str, ri->recType));              // ???? -slot?
    TAP_Print("\nbyte_01             = %02x", ri->byte_01);              //22,68,          tuner     trickmode
    TAP_Print("\nbyte_02             = %02x", ri->byte_02);              //                rsvd      speed
    TAP_Print("\nbyte_03             = %02x", ri->byte_03);              //                svcType
    TAP_Print("\nduration            = %d seconds", ri->duration);             //
    TAP_Print("\nchNum               = %d",   ri->chNum);                //chnum     ?
    TAP_Print("\nstart time          = %s",   mjdToString(ri->start_time));
    TAP_Print("\nend time            = %s",   mjdToString(ri->end_time));
    TAP_Print("\nword_10             = %04x", ri->word_10);              //chtype?
    TAP_Print("\nservice_id          = %04x", ri->service_id);           //
    TAP_Print("\nfilename            = %s",   ri->filename);             //
    TAP_Print("\ndword_54            = %08x", ri->dword_54);             //
    TAP_Print("\ndword_58            = %08x", ri->dword_58);             //
    TAP_Print("\ndword_5c            = %08x", ri->dword_5c);             //
    TAP_Print("\ndword_60            = %08x", ri->dword_60);             //
    TAP_Print("\nword_64             = %04x", ri->word_64);              //
    TAP_Print("\npmt_pid             = %04x", ri->pmt_pid);              //
    TAP_Print("\ndword_68            = %08x", ri->dword_68);             //
    TAP_Print("\ndword_6c            = %08x", ri->dword_6c);             //
    TAP_Print("\ndword_70            = %08x", ri->dword_70);             //
    TAP_Print("\ndword_74            = %08x", ri->dword_74);             //
    TAP_Print("\nuhf_channel         = %d",   ri->uhf_channel);          //
    TAP_Print("\nbandwidth           = %d MHz", ri->bandwidth);          //
    TAP_Print("\nbyte_7b             = %02x", ri->byte_7b);              //00
    TAP_Print("\nfrequency           = %d MHz", ri->frequency);          //
    TAP_Print("\ntransport_stream_id = %04x", ri->transport_stream_id);  //
    TAP_Print("\nword_82             = %04x", ri->word_82);              //0000 ???
    TAP_Print("\noriginal_network_id = %04x", ri->original_network_id);  //
    TAP_Print("\nnetwork_id          = %04x", ri->network_id);           //3005 ???
}

void dump_ci(type_channel_info *ci)
{
    TAP_Print("\nservice_id          = %04x", ci->service_id);
    TAP_Print("\nvideo_pid           = %04x", ci->video_pid);
    TAP_Print("\naudio_pid           = %04x", ci->audio_pid);
    TAP_Print("\npcr_pid             = %04x", ci->pcr_pid);
    TAP_Print("\ntransport_stream_id = %04x", ci->transport_stream_id);
    TAP_Print("\noriginal_network_id = %04x", ci->original_network_id);
    TAP_Print("\npmt_pid             = %04x", ci->pmt_pid);
    TAP_Print("\nword_0e             = %04x", ci->word_0e);     //uhf_channel?
    TAP_Print("\nbyte_10             = %02x", ci->byte_10);
    TAP_Print("\nbyte_11             = %02x", ci->byte_11);
    TAP_Print("\nword_12             = %04x", ci->word_12);
}


int dump_recording_status()
{
    char str[0x100];
    dword dwRgn, dwCol, dwRow, dwWidth, dwHeight, dwMaxCol;

    int slot;
    TYPE_RecInfo ri[2];
    byte dummy3[0x80];
    type_channel_info ci;
    TYPE_TapChInfo chInfo[3];
    int i;
    int emma_pid, active, slot0, slot1, pid;

    int channel_number;
    int ret;


    //delete any old ones
    if (recording_status_showing) {
        TAP_Win_Delete(&recording_status_window);
        TAP_EnterNormal();
        recording_status_showing = 0;
    }

    dwWidth = 500;
    dwCol = ((720 - dwWidth) / 2) - 25;
    dwRow = 100;
    dwHeight = 300;
    dwMaxCol = dwCol + 18 + dwWidth + 18;

    TAP_ExitNormal();
    TAP_Win_SetDefaultColor(&recording_status_window);
    TAP_Win_Create(&recording_status_window, recording_status_rgn, dwCol, dwRow, dwWidth, dwHeight, TRUE, TRUE);
    TAP_Win_SetTitle(&recording_status_window, "Recording Status", 1, FNT_Size_1622);
    dwRow = 150;
    recording_status_showing = 1;
    TAP_Win_SetFont(&recording_status_window, ALIGN_LEFT, FNT_Size_1622);


    slot = (int)*fw_tshift_slot;

    for (i=0; i<2; i++) {
        TAP_Hdd_GetRecInfo(i, &ri[i]);
        if (ri[i].recType) {
            TAP_Channel_GetInfo(ri[i].svcType, ri[i].svcNum, &chInfo[i]);
            if (ri[i].recType == RECTYPE_Normal) {
                sprintf(str, "Slot %d Recording: %3d - %s", i, chInfo[i].logicalChNum, chInfo[i].chName);
                TAP_Win_AddItem(&recording_status_window, str);
                dwRow += 20;
                sprintf(str, "  Video: %04x, Audio: %04x, PCR: %04x, PMT: %04x", chInfo[i].videoPid, 0x1fff & chInfo[i].audioPid, chInfo[i].pcrPid, chInfo[i].pmtPid);
                TAP_Win_AddItem(&recording_status_window, str);
                dwRow += 20;
            }
            if (ri[i].recType == RECTYPE_Timeshift) {
                sprintf(str, "Slot %d Timeshifting: %3d - %s", i, chInfo[i].logicalChNum, chInfo[i].chName);
                TAP_Win_AddItem(&recording_status_window, str);
                dwRow += 20;
                sprintf(str, "  Video: %04x, Audio: %04x, PCR: %04x, PMT: %04x", chInfo[i].videoPid, 0x1fff & chInfo[i].audioPid, chInfo[i].pcrPid, chInfo[i].pmtPid);
                TAP_Win_AddItem(&recording_status_window, str);
                dwRow += 20;
            }
        }
    }


    //check what is recording
    for (i=0; i<32; i++) {
        emma_pid = *(dword *)(EMMA_TTSD_SR_PID0+(i*4));
        active = (emma_pid & 0x00020000) ? 1 : 0;
        slot0 = (emma_pid & 0x00010000) ? 1 : 0;
        slot1 = (emma_pid & 0x00002000) ? 1 : 0;
        pid = emma_pid & 0x1fff;
        if (active) {

            if (srpid[i].recType != RECTYPE_None) {
//                if (srpid[i].recType != 3) {
                if (srpid[i].channel_number >= 0) {
                    TAP_Channel_GetInfo(srpid[i].svcType, srpid[i].channel_number, &chInfo[2]);
                    sprintf(str, "%2d: %s slot=%x, pid=%04x, %d.%s",
                        i,
                        (srpid[i].recType == RECTYPE_Normal) ? "Record" : (srpid[i].recType == RECTYPE_Timeshift) ? "TShift" : (srpid[i].recType == 3) ? "firmw " : "???   ",
                        srpid[i].slot, pid, chInfo[2].logicalChNum, chInfo[2].chName
                    );
                }
                else {
//                    sprintf(str, "%2d: %s slot=%x, pid=%04x, ---firmware---",
                    sprintf(str, "%2d: %s slot=%x, pid=%04x, %d.%s",
                        i,
                        (srpid[i].recType == RECTYPE_Normal) ? "Record" : (srpid[i].recType == RECTYPE_Timeshift) ? "TShift" : (srpid[i].recType == 3) ? "firmw " : "???   ",
//                        srpid[i].slot, pid
                        srpid[i].slot, pid, chInfo[slot1].logicalChNum, chInfo[slot1].chName
                    );
                }
                TAP_Print("\n%s", str);
                TAP_Win_AddItem(&recording_status_window, str);
                dwRow += 20;
            }
            else {
                TAP_Print("\n***oops, user srpid=0 *** %8x : slot0 = %d, slot1 = %d, EMMA_TTSD_SR_PID%-2d = %04x", emma_pid, slot0, slot1, i, pid);
             }
        }
        else {
            if (srpid[i].recType != RECTYPE_None) {
                TAP_Print("\n***oops, fw srpid=0 *** SR %d", i);
            }
            else {
            }
        }
    }
}

int dump_playing_status()
{
    char str[0x100];
    char str1[0x100];
    dword dwRgn, dwCol, dwRow, dwWidth, dwHeight, dwMaxCol;

    int channel_type, channel_number;
    int emma_ch, source;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    TYPE_TapChInfo ci;

    //delete any old ones
    if (playing_status_showing) {
        TAP_Win_Delete(&playing_status_window);
        TAP_EnterNormal();
        playing_status_showing = 0;
    }

    dwWidth = 500;
    dwCol = ((720 - dwWidth) / 2) - 25;
    dwRow = 100;
    dwHeight = 300;
    dwMaxCol = dwCol + 18 + dwWidth + 18;

    TAP_ExitNormal();
    TAP_Win_SetDefaultColor(&playing_status_window);
    TAP_Win_Create(&playing_status_window, playing_status_rgn, dwCol, dwRow, dwWidth, dwHeight, TRUE, TRUE);
    TAP_Win_SetTitle(&playing_status_window, "Playback Status", 1, FNT_Size_1622);
    dwRow = 150;
    playing_status_showing = 1;
    TAP_Win_SetFont(&playing_status_window, ALIGN_LEFT, FNT_Size_1622);


    emma_ch = (*(dword *)(EMMA_TTSD_1ST_VIDEO_PID) & 0x60000000) >> 29;
    source = demap_channel(emma_ch);
    if ((source == 0) || (source == 1)) {
        sprintf(str, "Main Channel, using tuner %d", source);
        TAP_Channel_GetCurrent(&channel_type, &channel_number);
        TAP_Channel_GetInfo(channel_type, playing_chNum, &ci);
        sprintf(str1, ": %d. %s", ci.logicalChNum, ci.chName);
        strcat(str, str1);
    }
    else if (source == 2) {
        sprintf(str, "Main Channel, using file");
        TAP_Hdd_GetPlayInfo(&pi);
        if ((pi.playMode == PLAYMODE_Playing) || (pi.playMode != PLAYMODE_TempPlaying) || (pi.playMode != PLAYMODE_RecPlaying)) {
            TAP_Channel_GetInfo(pi.svcType, playing_chNum, &ci);
            sprintf(str1, ": %d. %s", ci.logicalChNum, ci.chName);
            strcat(str, str1);
        }
    }
    else {
        sprintf(str, "Main Channel, using ???");
    }
    TAP_Win_AddItem(&playing_status_window, str);
    sprintf(str, "    PCR PID = %04x", 0x1fff & *(dword *)(EMMA_TTSD_1ST_PCR_PID));
    TAP_Win_AddItem(&playing_status_window, str);
    sprintf(str, "    VIDEO PID = %04x", 0x1fff & *(dword *)(EMMA_TTSD_1ST_VIDEO_PID));
    TAP_Win_AddItem(&playing_status_window, str);
    sprintf(str, "    AUDIO PID = %04x", 0x1fff & *(dword *)(EMMA_TTSD_1ST_AUDIO_PID));
    TAP_Win_AddItem(&playing_status_window, str);




    emma_ch = (*(dword *)(EMMA_TTSD_2ND_VIDEO_PID) & 0x60000000) >> 29;
    source = demap_channel(emma_ch);
    if ((source == 0) || (source == 1)) {
        sprintf(str, "Sub Channel, using tuner %d", source);
        TAP_Channel_GetInfo(SVC_TYPE_Tv, sub_chNum, &ci);
        sprintf(str1, ": %d. %s", ci.logicalChNum, ci.chName);
        strcat(str, str1);
    }
    else if (source == 2) {
        sprintf(str, "Sub Channel, using file");
        TAP_Hdd_GetPlayInfo(&pi);
        if ((pi.playMode == PLAYMODE_Playing) || (pi.playMode != PLAYMODE_TempPlaying) || (pi.playMode != PLAYMODE_RecPlaying)) {
            TAP_Channel_GetInfo(pi.svcType, sub_chNum, &ci);
            sprintf(str1, ": %d. %s", ci.logicalChNum, ci.chName);
            strcat(str, str1);
        }
    }
    else {
        sprintf(str, "Sub Channel, using ???");
    }
    TAP_Win_AddItem(&playing_status_window, str);
    sprintf(str, "    PCR PID   = %04x", 0x1fff & *(dword *)(EMMA_TTSD_2ND_PCR_PID));
    TAP_Win_AddItem(&playing_status_window, str);
    sprintf(str, "    VIDEO PID = %04x", 0x1fff & *(dword *)(EMMA_TTSD_2ND_VIDEO_PID));
    TAP_Win_AddItem(&playing_status_window, str);
    sprintf(str, "    AUDIO PID = %04x", 0x1fff & *(dword *)(EMMA_TTSD_2ND_AUDIO_PID));
    TAP_Win_AddItem(&playing_status_window, str);
}
